/* NiuTrans.Tensor - an open-source tensor library
* Copyright (C) 2018, Natural Language Processing Lab, Northestern University.
* All rights reserved.
*
* Licensed under the Apache License, Version 2.0 (the "License");
* you may not use this file except in compliance with the License.
* You may obtain a copy of the License at
*
*   http://www.apache.org/licenses/LICENSE-2.0
*
* Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an "AS IS" BASIS,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the specific language governing permissions and
* limitations under the License.
*/

/*
*
* This is a simple impelementation of the feed-forward network-baesd language
* model (FNNLM). See more details about FNNLM in
* "A Neural Probabilistic Language Model" by Bengio et al.
* Journal of Machine Learning Research 3 (2003) 1137C1155
*
* $Created by: ZHANG Yuhao (yoohao.zhang@gmail.com) 2018-08-08
*/

#include "myFNNLM.h"
#include "../../tensor/XGlobal.h"
#include "../../tensor/XUtility.h"
#include "../../tensor/XDevice.h"
#include "../../tensor/function/FHeader.h"
#include "../../network/XNet.h"
#include <cuda_runtime.h>

namespace myfnnlm
{
#define Vocabulary_Size 16400
#define m_size 100
#define h_size 256
#define devID 0
#define LR 0.001
DTYPE xData[Vocabulary_Size * h_size];

void shape(XTensor *t)
{
    int dim = t->order;
    printf("dim is %d: ", dim);
    for (int i = 0; i < dim; ++i)
    {
        printf("%d ,", t->dimSize[i]);
    }
    printf("\n");
}

void show(XTensor *a)
{
    float data[50];
    cudaMemcpy(data, a->data, 50 * sizeof(float), cudaMemcpyDeviceToHost);
    for (int i = 0; i < 50; ++i)
    {
        printf("%f ", data[i]);
    }
    printf("\n");
}
void show(XTensor *a,int num)
{
    float data[200];
    cudaMemcpy(data, a->data, num * sizeof(float), cudaMemcpyDeviceToHost);
    for (int i = 0; i < num; ++i)
    {
        printf("%f ", data[i]);
    }
    printf("\n");
}

FILE* read_file_line(FILE* file ,int *data,int* dataNum)
{
    if (file == NULL)
    {
        printf("error");
        exit(1);
    }
    else
    {
        char string[1024];
        *dataNum = 0;
        fgets(string, 1024, file);
        int flag = 0;
        int start_index = 0;
        for (int i = 0; i < strlen(string); ++i)
        {
            if ('0' <= string[i] && string[i] <= '9')
            {
                if (!flag)
                {
                    flag = 1;
                    start_index = i;
                }
            }
            else
            {
                if (flag)
                {
                    int tmp = 0;
                    for (int j = start_index; j < i; ++j)
                    {
                        tmp = tmp * 10 + string[j] - '0';
                    }
                    data[(*dataNum)++] = tmp;
                    flag = 0;
                }
            }
        }
        if (flag)
        {
            int tmp = 0;
            for (int j = start_index; j < strlen(string); ++j)
            {
                tmp = tmp * 10 + string[j] - '0';
            }
            data[(*dataNum)++] = tmp;
            flag = 0;
        }
    }
    return file;//update the file point
}

FILE* read_file_number(FILE* file, int *data, int n_gram,int batch, int* true_batch,int* head)
{
    int cot = 0;
    bool sentence_start = false;
    if (head[0] == -1)
    {
        for (int i = 0; i < n_gram - 1; ++i)
            fscanf(file, "%d", &data[cot++]);
    }
    else
    {
        int tmp;
        fscanf(file, "%d", &tmp);
        if (tmp == 1)
            sentence_start = true;
        else {
            for (int i = 0; i < n_gram - 1; ++i)
            {
                data[cot++] = head[i];
            }
            data[cot++] = tmp;
            batch--;
        }
    }
    while (batch && !feof(file))
    {
        if (sentence_start)
        {
            for (int i = 0; i < n_gram - 1; ++i)
                fscanf(file, "%d", &data[cot++]);
            sentence_start = false;
        }
        else
        {
            fscanf(file, "%d", &data[cot]);
            if (data[cot] == 1)
                sentence_start = true;
            else batch--;
            cot++;
        }    
    }
    for (int i = cot - n_gram + 1,j=0; i < cot; ++i,++j)
    {
        head[j] = data[i];
    }
    *true_batch = batch;
    return file;
}

FILE* write_tensor(XTensor* t, FILE* file)
{
    fprintf(file, "%d :", t->order);
    for (int i = 0; i < t->order; ++i)
    {
        fprintf(file, "%d ", t->dimSize[i]);
    }
    fprintf(file, "\n");
    int num = t->unitNum;
    float* data = new float[num];
    cudaMemcpy(data, t->data, (t->unitNum) * sizeof(float), cudaMemcpyDeviceToHost);
    for (int i = 0; i < num - 1; ++i)
    {
        fprintf(file, "%e ", data[i]);
        //printf("%f ", data[i]);
    }
    fprintf(file, "%e", data[num - 1]);
    fprintf(file,"\n");
    delete[] data;
    return file;
}

float cot_loss(XTensor *output,XTensor *right,int batch)
{
    int order = 3;
    int* dimSize = new int[order];
    dimSize[0] = batch;
    dimSize[1] = 1;
    dimSize[2] = 1;
    XTensor* batch_ans_tensor = NewTensor(order, dimSize, X_FLOAT, 1.0F, devID);
    dimSize[0] = batch;
    dimSize[1] = Vocabulary_Size;
    dimSize[2] = 1;
    output->Reshape(order, dimSize);
    right->Reshape(order, dimSize);

    _MatrixMulBatched(output, X_TRANS, right, X_NOTRANS, batch_ans_tensor);

    order = 2;
    dimSize[0] = 1;
    dimSize[1] = 1;
    XTensor* ans_tensor = NewTensor(order, dimSize, X_FLOAT, 1.0F, devID);
    _ReduceSum(batch_ans_tensor, ans_tensor, 0);
    float loss = ans_tensor->Get2D(0,0);

    dimSize[0] = batch;
    dimSize[1] = Vocabulary_Size;
    output->Reshape(order, dimSize);
    right->Reshape(order, dimSize);

    delete[] dimSize;
    return loss;
}

void train(int n_gram, int batch, int nEpoch, bool write_model)
{
    char file_path[200] = "brown/brown.train.5gram";
    //char file_path[200] = "brown/test.txt";
    FILE *file;
    file = fopen(file_path, "rb");
    int* data = new int[(n_gram - 1) * batch + 5];
    int* head = new int[n_gram - 1];
    memset(head, -1, sizeof(int)*(n_gram - 1));

    for (int i = 0; i < Vocabulary_Size * h_size; ++i)
    {
        xData[i] = 0.1;
    }
    float learn_rate = LR;
    int order = 3;
    int* dimSize = new int[order];
    order = 2;
    dimSize[0] = Vocabulary_Size;
    dimSize[1] = m_size;
    XTensor *feature_vector = NewTensor(order, dimSize, X_FLOAT, 1.0F, devID);//V * m tensor
    feature_vector->SetDataRand(-0.1F, 0.1F);
    //feature_vector->SetData(xData, feature_vector->unitNum);

    order = 2;
    dimSize[0] = m_size * (n_gram - 1);
    dimSize[1] = h_size;
    XTensor *H = NewTensor(order, dimSize, X_FLOAT, 1.0F, devID);//H m(n-1) * h tensor
    XTensor *update_H = NewTensor(order, dimSize, X_FLOAT, 1.0F, devID);//for update H m(n-1) * h tensor
    H->SetDataRand(-0.1F, 0.1F);
    //H->SetData(xData, H->unitNum);

    order = 1;
    dimSize[0] = h_size;
    XTensor *d = NewTensor(order, dimSize, X_FLOAT, 1.0F, devID);//d h vector
    XTensor *update_d = NewTensor(order, dimSize, X_FLOAT, 1.0F, devID);//update d h vector
    d->SetDataRand(-0.1F, 0.1F);
    //d->SetData(xData, d->unitNum);

    order = 2;
    dimSize[0] = h_size;
    dimSize[1] = Vocabulary_Size;
    XTensor *U = NewTensor(order, dimSize, X_FLOAT, 1.0F, devID);//U h * V tensor
    XTensor *update_U = NewTensor(order, dimSize, X_FLOAT, 1.0F, devID);//update U h * V tensor
    U->SetDataRand(-0.1F, 0.1F);
    //float tmp_data[h_size * Vocabulary_Size];
    //for (int i = 0; i < U->unitNum; ++i)
    //{
    //    tmp_data[i] = i / h_size;
    //}
    //U->SetData(tmp_data, U->unitNum);

    order = 1;
    dimSize[0] = Vocabulary_Size;
    XTensor *b = NewTensor(order, dimSize, X_FLOAT, 1.0F, devID);//b V * 1 tensor
    XTensor *update_b = NewTensor(order, dimSize, X_FLOAT, 1.0F, devID);//for update V * 1 tensor
    b->SetDataRand(-0.1F, 0.1F);
    //b->SetData(xData, b->unitNum);

    /* hide node */
    order = 3;
    dimSize[0] = batch;
    dimSize[1] = n_gram - 1;
    dimSize[2] = Vocabulary_Size;
    XTensor *onehot = NewTensor(order, dimSize, X_FLOAT, 1.0F, devID);

    order = 3;
    dimSize[0] = batch;
    dimSize[1] = n_gram - 1;
    dimSize[2] = m_size;
    XTensor *embedding = NewTensor(order, dimSize, X_FLOAT, 1.0F, devID);

    order = 2;
    dimSize[0] = batch;
    dimSize[1] = h_size;
    XTensor *hide_node = NewTensor(order, dimSize, X_FLOAT, 1.0F, devID);
    XTensor *hide_node2 = NewTensor(order, dimSize, X_FLOAT, 1.0F, devID);
    XTensor *hide_node3 = NewTensor(order, dimSize, X_FLOAT, 1.0F, devID);

    dimSize[0] = batch;
    dimSize[1] = Vocabulary_Size;
    XTensor *hide_node4 = NewTensor(order, dimSize, X_FLOAT, 1.0F, devID);
    XTensor *hide_node5 = NewTensor(order, dimSize, X_FLOAT, 1.0F, devID);
    XTensor *output = NewTensor(order, dimSize, X_FLOAT, 1.0F, devID);

    dimSize[0] = batch;
    dimSize[1] = Vocabulary_Size;
    XTensor* gold = NewTensor(order, dimSize, X_FLOAT, 1.0F, devID);

    /*backward node*/
    order = 2;
    dimSize[0] = batch;
    dimSize[1] = Vocabulary_Size;
    XTensor* ded_hide_node5 = NewTensor(order, dimSize, X_FLOAT, 1.0F, devID);

    dimSize[0] = h_size;
    dimSize[1] = Vocabulary_Size;
    XTensor* ded_hide_node3 = NewTensor(order, dimSize, X_FLOAT, 1.0F, devID);

    dimSize[0] = batch;
    dimSize[1] = h_size;
    XTensor* gradPassed = NewTensor(order, dimSize, X_FLOAT, 1.0F, devID);

    dimSize[0] = batch;
    dimSize[1] = h_size;
    XTensor* ded_hide_node2 = NewTensor(order, dimSize, X_FLOAT, 1.0F, devID);

    dimSize[0] = m_size * (n_gram - 1);
    dimSize[1] = h_size;
    XTensor* ded_hide_node = NewTensor(order, dimSize, X_FLOAT, 1.0F, devID);

    dimSize[0] = batch;
    dimSize[1] = m_size * (n_gram - 1);
    XTensor* ded_embedding = NewTensor(order, dimSize, X_FLOAT, 1.0F, devID);

    order = 3;
    dimSize[0] = batch;
    dimSize[1] = Vocabulary_Size;
    dimSize[2] = m_size;
    XTensor* ded_feature_vector = NewTensor(order, dimSize, X_FLOAT, 1.0F, devID);

    XList fList(batch);
    order = 2;
    dimSize[0] = Vocabulary_Size;
    dimSize[1] = m_size;
    for (int i = 0; i < batch; ++i)
    {
        XTensor* update_feature_vector = NewTensor(order, dimSize, X_FLOAT, 1.0F, devID);//for update V * m tensor
        fList.Add(update_feature_vector);
    }
    for(int k = 0; k < nEpoch; ++k)
    {
        int total_Num = 0;
        float loss = 0;
        int cot = 0;
        while (!feof(file))
        {
            //file = read_file_line(file, data, &dataNum);
            int true_batch = batch;
            file = read_file_number(file, data, n_gram, batch, &true_batch, head);
            true_batch = batch - true_batch;
            int* answer = new int[true_batch];
            total_Num += true_batch;
            onehot->SetZeroAll();
            int index = 0;
            for (int i = 0; i < true_batch; ++i)
            {
                if (data[index + n_gram - 1] == 1)
                {
                    index = index + n_gram;
                }
                answer[i] = data[index + n_gram - 1];
                for (int j = index; j < index + n_gram - 1; ++j)
                {
                    onehot->Set3D(1.0F, i, j - index, data[j]);
                }
                index++;
            }
            order = 3;
            dimSize[0] = batch;
            dimSize[1] = n_gram - 1;
            dimSize[2] = m_size;
            embedding->Reshape(order, dimSize);

            _MatrixMul(onehot, X_NOTRANS, feature_vector, X_NOTRANS, embedding);

            order = 2;
            dimSize[0] = batch;
            dimSize[1] = (n_gram - 1) * m_size;
            embedding->Reshape(order, dimSize);//batch x m(n-1) input tensor

            _MatrixMul(embedding, X_NOTRANS, H, X_NOTRANS, hide_node);//batch x h

            _SumDim(hide_node, d, hide_node2, 1);//batch x h

            _HardTanH(hide_node2, hide_node3);//batch x h

            _MatrixMul(hide_node3, X_NOTRANS, U, X_NOTRANS, hide_node4);//batch x V

            _SumDim(hide_node4, b, hide_node5, 1);//batch x V

            _LogSoftmax(hide_node5, output, 1);

            /*order = 2;
            dimSize[0] = batch;
            dimSize[1] = Vocabulary_Size;
            gold->Reshape(order,dimSize);*/
            gold->SetZeroAll();
            //printf("answer is %d ", answer);
            for (int i = 0; i < true_batch; ++i)
            {
                gold->Set2D(1.0F, i, answer[i]);
            }
            float test = cot_loss(output, gold, batch);
            loss += -test;
            //printf("prob is %f \n", test);

            //backward
            /*order = 2;
            dimSize[0] = batch;
            dimSize[1] = Vocabulary_Size;
            gold->Reshape(order,dimSize);*/

            _LogSoftmaxBackward(gold, output, hide_node5, NULL, ded_hide_node5, 1, CROSSENTROPY);

            _MatrixMul(hide_node3, X_TRANS, ded_hide_node5, X_NOTRANS, ded_hide_node3);
            _MatrixMul(ded_hide_node5, X_NOTRANS, U, X_TRANS, gradPassed);

            _HardTanHBackward(NULL, hide_node3, hide_node2, gradPassed, ded_hide_node2, NOLOSS);

            order = 2;
            dimSize[0] = batch;
            dimSize[1] = m_size * (n_gram - 1);
            ded_embedding->Reshape(order, dimSize);
            _MatrixMul(ded_hide_node2, X_NOTRANS, H, X_TRANS, ded_embedding);

            order = 3;
            dimSize[0] = batch;
            dimSize[1] = n_gram - 1;
            dimSize[2] = m_size;
            ded_embedding->Reshape(order, dimSize);
            _MatrixMulBatched(onehot, X_TRANS, ded_embedding, X_NOTRANS, ded_feature_vector);

            //update
            //_ScaleAndShiftMe(ded_hide_node5,learn_rate);
            _ReduceSum(ded_hide_node5, update_b, 0);
            _SumMe(b, update_b, -learn_rate);//update b

            //_ReduceSum(ded_hide_node3, update_U, 0);
            //_SumDim(U, ded_hide_node3, 0);
            _SumMe(U, ded_hide_node3, -learn_rate);//update U

            _ReduceSum(ded_hide_node2, update_d, 0);
            _SumMe(d, update_d, -learn_rate);//update d

            _MatrixMul(embedding, X_TRANS, ded_hide_node2, X_NOTRANS, ded_hide_node);
            //_ReduceSum(ded_hide_node, update_H, 0);
            _SumMe(H, ded_hide_node, -learn_rate);//update H

            _Split(ded_feature_vector, &fList, 0, batch);
            //_ReduceSum(ded_feature_vector,update_feature_vector,0);
            for (int i = 0; i < batch; ++i)
            {
                XTensor* update_feature_vector = (XTensor*)fList.GetItem(i);
                _SumMe(feature_vector, update_feature_vector, -learn_rate);
            }
            cot ++ ;
            if (cot % 200 == 0) {
                printf("%d %f\n", cot, exp(loss / total_Num));
                //printf("%f %d\n", loss, total_Num);
            }
            delete[] answer;
        }
    }
    if (write_model)
    {
        char out_file_path[200] = "output.txt";
        FILE *out_file;
        out_file = fopen(out_file_path, "wb");
        if (out_file != NULL)
        {
            out_file = write_tensor(feature_vector,out_file);
            out_file = write_tensor(H, out_file);
            out_file = write_tensor(d, out_file);
            out_file = write_tensor(U, out_file);
            out_file = write_tensor(b, out_file);
        }
        else
        {
            printf("can't open model output file\n");
        }
        fclose(out_file);
    }
    delete[] data;
    delete[] dimSize;
    delete[] head;
    fclose(file);
}

void forward(int n_gram, int batch)
{
    char file_path[200] = "brown/brown.test.5gram";
    //char file_path[200] = "brown/test.txt";
    FILE *file;
    file = fopen(file_path, "rb");
    int* data = new int[(n_gram - 1) * batch +5];
    int* head = new int[n_gram - 1];
    memset(head, -1, sizeof(int)*(n_gram - 1));

    //char file_path1[200] = "model_parameters.out";
    char file_path1[200] = "output.txt";
    FILE *model_file;
    model_file = fopen(file_path1, "rb");
    if (model_file == NULL)
    {
        printf("file open error\n");
        exit(1);
    }
    char string[1024];
    fgets(string, 1024, model_file);
    int order = 3;
    int* dimSize = new int[order];
    order = 2;
    dimSize[0] = Vocabulary_Size;
    dimSize[1] = 100;
    XTensor *feature_vector = NewTensor(order, dimSize, X_FLOAT, 1.0F, devID);//V * m tensor

    char dataString[32];
    for (int i = 0; i <Vocabulary_Size * 100; ++i)
    {
        if (fscanf(model_file, "%s", dataString) != EOF)
        {
            //printf("%s", dataString);
            xData[i] = atof(dataString);
            //srcTensorData[i] = i;
        }
        else
        {
            printf("read wrong");
            break;
        }
    }

    feature_vector->SetData(xData, Vocabulary_Size * 100);
    fscanf(model_file, "%c", string);//filter the '\n'
    fgets(string, 1024, model_file);
    for (int i = 0; i <100 * (n_gram - 1) * 256; ++i)
    {
        if (fscanf(model_file, "%s", dataString) != EOF)
        {
            //printf("%s", dataString);
            xData[i] = atof(dataString);
            //srcTensorData[i] = i;
        }
        else
        {
            printf("read wrong");
            break;
        }
    }

    order = 2;
    dimSize[0] = 100 * (n_gram - 1);
    dimSize[1] = 256;
    XTensor *H = NewTensor(order, dimSize, X_FLOAT, 1.0F, devID);//H m(n-1) * h tensor
    H->SetData(xData, 256 * 100 * (n_gram - 1));
    fscanf(model_file, "%c", string);//filter the '\n'
    fgets(string, 1024, model_file);
    for (int i = 0; i < 256; ++i)
    {
        if (fscanf(model_file, "%s", dataString) != EOF)
        {
            //printf("%s", dataString);
            xData[i] = atof(dataString);
            //srcTensorData[i] = i;
        }
        else
        {
            printf("read wrong");
            break;
        }
    }

    order = 2;
    dimSize[0] = 1;
    dimSize[1] = 256;
    XTensor *d = NewTensor(order, dimSize, X_FLOAT, 1.0F, devID);//d h vector
    d->SetData(xData, 256);

    fscanf(model_file, "%c", string);//filter the '\n'
    fgets(string, 1024, model_file);
    for (int i = 0; i < 256 * Vocabulary_Size; ++i)
    {
        if (fscanf(model_file, "%s", dataString) != EOF)
        {
            //printf("%s", dataString);
            xData[i] = atof(dataString);
            //srcTensorData[i] = i;
        }
        else
        {
            printf("read wrong");
            break;
        }
    }

    order = 2;
    dimSize[0] = 256;
    dimSize[1] = Vocabulary_Size;
    XTensor *U = NewTensor(order, dimSize, X_FLOAT, 1.0F, devID);//U h * V tensor
    U->SetData(xData, 256 * Vocabulary_Size);

    fscanf(model_file, "%c", string);//filter the '\n'
    fgets(string, 1024, model_file);
    for (int i = 0; i < Vocabulary_Size; ++i)
    {
        if (fscanf(model_file, "%s", dataString) != EOF)
        {
            //printf("%s", dataString);
            xData[i] = atof(dataString);
            //srcTensorData[i] = i;
        }
        else
        {
            printf("read wrong");
            break;
        }
    }

    order = 2;
    dimSize[0] = 1;
    dimSize[1] = Vocabulary_Size;
    XTensor *b = NewTensor(order, dimSize, X_FLOAT, 1.0F, devID);//b V * 1 tensor
    b->SetData(xData, Vocabulary_Size);

    /* hide node and output*/
    order = 3;
    dimSize[0] = batch;
    dimSize[1] = n_gram - 1;
    dimSize[2] = Vocabulary_Size;
    XTensor *onehot = NewTensor(order, dimSize, X_FLOAT, 1.0F, devID);

    order = 3;
    dimSize[0] = batch;
    dimSize[1] = n_gram - 1;
    dimSize[2] = 100;
    XTensor *embedding = NewTensor(order, dimSize, X_FLOAT, 1.0F, devID);

    order = 2;
    dimSize[0] = batch;
    dimSize[1] = 256;
    XTensor *hide_node = NewTensor(order, dimSize, X_FLOAT, 1.0F, devID);
    XTensor *hide_node2 = NewTensor(order, dimSize, X_FLOAT, 1.0F, devID);
    XTensor *hide_node3 = NewTensor(order, dimSize, X_FLOAT, 1.0F, devID);

    dimSize[0] = batch;
    dimSize[1] = Vocabulary_Size;
    XTensor *hide_node4 = NewTensor(order, dimSize, X_FLOAT, 1.0F, devID);
    XTensor *hide_node5 = NewTensor(order, dimSize, X_FLOAT, 1.0F, devID);
    XTensor *output = NewTensor(order, dimSize, X_FLOAT, 1.0F, devID);

    dimSize[0] = batch;
    dimSize[1] = Vocabulary_Size;
    XTensor* gold = NewTensor(order, dimSize, X_FLOAT, 1.0F, devID);

    float loss = 0;
    int total_Num = 0;
    int cot = 0;
    while (!feof(file))
    {
        cot++;
        int true_batch = batch;
        file = read_file_number(file, data, n_gram, batch, &true_batch, head);
        true_batch = batch - true_batch;
        total_Num += true_batch;
        //printf("%d \n", true_batch);
        onehot->SetZeroAll();
        int index = 0;
        int* answer = new int[true_batch];
        for (int i = 0; i < true_batch; ++i)
        {
            if (data[index + n_gram - 1] == 1)
            {
                index = index + n_gram;
            }
            answer[i] = data[index + n_gram - 1];
            for (int j = index; j < index + n_gram - 1; ++j)
            {
                //onehot->Set2D(1.0F, j - i, data[j]);
                onehot->Set3D(1.0F, i, j - index, data[j]);
                //printf("%d %d %d\n", i, j - index, data[j]);
            }
            //printf("\n");
            index++;
        }

        order = 3;
        dimSize[0] = batch;
        dimSize[1] = n_gram - 1;
        dimSize[2] = 100;
        embedding->Reshape(order, dimSize);

        _MatrixMul(onehot, X_NOTRANS, feature_vector, X_NOTRANS, embedding);

        order = 2;
        dimSize[0] = batch;
        dimSize[1] = (n_gram - 1) * 100;
        embedding->Reshape(order, dimSize);//batch x m(n-1) input tensor

        _MatrixMul(embedding, X_NOTRANS, H, X_NOTRANS, hide_node);//batch x h


        _SumDim(hide_node, d, hide_node2,1);//batch x h

        _HardTanH(hide_node2, hide_node3);//batch x h

        _MatrixMul(hide_node3, X_NOTRANS, U, X_NOTRANS, hide_node4);//batch x V


        _SumDim(hide_node4, b, hide_node5, 1);//batch x V

        _LogSoftmax(hide_node5, output, 1);//batch x V

        gold->SetZeroAll();
        for (int i = 0; i < true_batch; ++i)
        {
            gold->Set2D(1.0F,i, answer[i]);
        }
        float test = cot_loss(output, gold ,batch);
        //printf("prob is %f \n", test);
        loss += -test;
        if (cot % 100 == 0) printf("sentence %d  loss %f  num %d\n", cot, loss, total_Num);
        delete[] answer;
    }
    printf("%d \n", total_Num);
    printf("ppl is %f\n", exp(loss/ total_Num));
    delete[] dimSize;
    delete[] data;
    delete[] head;
    fclose(model_file);
    fclose(file);
}



void fnnlm()
{
    const int n_gram = 5;
    const int batch = 20;
    /*int data[batch * (n_gram - 1) + 5];
    int head[n_gram - 1] = { -1 };
    int number;
    char file_path[200] = "brown/brown.train.5gram";
    FILE *file;
    file = fopen(file_path, "rb");
    int cot = 0;
    while (!feof(file))
    {
        file = read_file_number(file, data, n_gram, batch, &number, head);
        //printf("%d ", number);
    }*/
    train(n_gram, batch, 2, true);
    /*int corpus_lenth = Vocabulary_Size;
    char file_path[200] = "brown/brown.test.5gram";
    FILE *file;
    file = fopen(file_path, "rb");
    if (file == NULL)
    {
        printf("error");
        exit(1);
    }
    else
    {
        char string[1024];
        while (!feof(file))
        {
            int sentence_data[150];
            int sentence_size = 0;
            fgets(string,1024,file);
            int flag = 0;
            int start_index = 0;
            for (int i = 0; i < strlen(string); ++i)
            {
                if ('0' <= string[i] && string[i] <= '9')
                {
                    if (!flag)
                    {
                        flag = 1;
                        start_index = i;
                    }
                }
                else
                {
                    if (flag)
                    {
                        int tmp = 0;
                        for (int j = start_index; j < i; ++j)
                        {
                            tmp = tmp * 10 + string[j] - '0';
                        }
                        sentence_data[sentence_size++] = tmp;
                        flag = 0;
                    }
                }
            }*/
            //forward(sentence_data,sentence_size,n_gram);
    forward(n_gram, batch);
        /*}
    }
    fclose(file);*/
}
}